/*
 * Copyright 2010 Red Hat, Inc.
 * Copyright 2012 Pete Zaitcev <zaitcev@yahoo.com>
 *
 * Test client for S3.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdbool.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <hstor.h>

enum poke_mode {
	MODE_NUL, MODE_IN, MODE_OUT, MODE_DEL, MODE_LIST
};

struct params {
	int mode;
	bool verbose;
	bool do_csum;
	bool subdomain;
	char *file;
	char *host;	/* netloc actually */
	char *user;
	char *pass;
	char *bucket;
	char *key;
};

struct input_ctx {
	struct params *par;
	struct hstor_client *httpstor;
	char *errbuf;
	char *pfx;
};

struct put_ctx {
	unsigned int csum;
	bool do_csum;
	int fd;
	off_t off;
	off_t total;
};

struct get_ctx {
	unsigned int csum;
	bool do_csum;
	FILE *f;
	off_t off;
};

#define BLKSZ	65536

#define CSUM_INIT  0xFFFFFFFF

static void do_input(struct params *par, struct hstor_client *httpstor, char *errbuf);
static void input_many(struct params *par, struct hstor_client *httpstor, char *errbuf);
static void input_func(gpointer data, gpointer priv);
static void input_one(struct params *par, struct hstor_client *httpstor,
    char *errbuf, char *key, FILE *f, char *filename);
static void do_output(struct params *par, struct hstor_client *httpstor, char *errbuf);
static void output_dir(struct params *par, struct hstor_client *httpstor,
    char *dirname, char *suffix);
static void output_name(struct params *par, struct hstor_client *httpstor,
    char *basekey, char *basedir, char *suffix);
static void output_one(struct params *par, struct hstor_client *httpstor,
    char *key, int fd, const char *filename);
static void do_delete(struct params *par, struct hstor_client *httpstor, char *errbuf);
static void do_list(struct params *par, struct hstor_client *httpstor, char *errbuf);
static size_t put_cb(void *ptr, size_t membsize, size_t nmemb, void *user_data);
static size_t get_cb(void *ptr, size_t membsize, size_t nmemb, void *user_data);
static char *cat_fs(char *dir, char *file);
static char *cat(int n, char *v[]);
static void mkpath(char *path);
static int isdir(char *filename);
static void incrsum(unsigned int *psum, const unsigned char *data, size_t len);
static void parse_args(struct params *par, int argc, char **argv);
static void Usage(void);

struct params pars;

int main(int argc, char **argv)
{
	struct hstor_client *httpstor;
	char *acc;
	char *errbuf;
	// bool rcb;
	int rc;

	parse_args(&pars, argc, argv);

	acc = pars.host;

	httpstor = hstor_new(acc, pars.host, pars.user, pars.pass);
	if (httpstor == NULL) {
		fprintf(stderr, "Unable to initialize a channel to %s\n", acc);
		exit(1);
	}

	if (pars.verbose)
		httpstor->verbose = true;
	if (pars.subdomain)
		hstor_set_format(httpstor, HFMT_SUBDOMAIN);

	if ((errbuf = malloc(CURL_ERROR_SIZE)) == NULL) {
		fprintf(stderr, "No core\n");
		exit(1);
	}
	strcpy(errbuf, "HAXX SCREWUP, USE -v");
	rc = curl_easy_setopt(httpstor->curl, CURLOPT_ERRORBUFFER, errbuf);
	if (rc) {
		fprintf(stderr, "curl_easy_setopt(CURLOPT_ERRORBUFFER): %d\n", rc);
	}

	switch (pars.mode) {
	case MODE_IN:
		do_input(&pars, httpstor, errbuf);
		break;
	case MODE_OUT:
		do_output(&pars, httpstor, errbuf);
		break;
	case MODE_DEL:
		do_delete(&pars, httpstor, errbuf);
		break;
	case MODE_LIST:
		do_list(&pars, httpstor, errbuf);
		break;
	default:
		abort();
	}

	hstor_free(httpstor);
	free(errbuf);
	return 0;
}

static void do_input(struct params *par, struct hstor_client *httpstor, char *errbuf)
{
	FILE *f;

	if (isdir(par->file)) {
		input_many(par, httpstor, errbuf);
		return;
	}

	if (par->file) {
		/* XXX open before opening connection to the server */
		if ((f = fopen(par->file, "wb")) == NULL) {
			fprintf(stderr, "Unable to open %s: %s\n",
			    par->file, strerror(errno));
			exit(1);
		}
		input_one(par, httpstor, errbuf, par->key, f, par->file);
		fclose(f);
	} else {
		input_one(par, httpstor, errbuf, par->key, stdout, "-");
	}
}

static void input_many(struct params *par, struct hstor_client *httpstor, char *errbuf)
{
	char *pfx;
	struct hstor_keylist *list;
	struct input_ctx inctx;

	pfx = par->key;
	list = hstor_keys(httpstor, par->bucket, pfx, NULL, NULL, 0);
	if (!list) {
		fprintf(stderr, "Failed to list for input, prefix `%s'\n", pfx);
		exit(1);
	}

	inctx.par = par;
	inctx.httpstor = httpstor;
	inctx.errbuf = errbuf;
	inctx.pfx = pfx;
	g_list_foreach(list->contents, input_func, &inctx);

	hstor_free_keylist(list);
}

static void input_func(gpointer data, gpointer priv)
{
	struct input_ctx *ctx = priv;
	struct hstor_object *obj = data;
	struct params *par = ctx->par;
	char *filename;
	FILE *f;

	/*
	 * XXX The list operation returns keys with common prefixes,
	 * so we create deeper hierarchies than we should, which do not
	 * match what output mode copied to the server. Need to pull prefixes.
	 *
	 * if (ctx->pfx && memcmp(ctx->pfx, obj->key, N) == 0) ...
	 */
	filename = cat_fs(par->file, obj->key);
	mkpath(filename);

	if ((f = fopen(filename, "wb")) == NULL) {
		fprintf(stderr, "Unable to open %s: %s\n",
		    filename, strerror(errno));
		exit(1);
	}
	input_one(par, ctx->httpstor, ctx->errbuf, obj->key, f, filename);
	fclose(f);

	free(filename);
}

static void input_one(struct params *par, struct hstor_client *httpstor,
    char *errbuf, char *key, FILE *f, char *filename)
{
	struct get_ctx getctx;

	memset(&getctx, 0, sizeof(getctx));
	getctx.f = f;
	getctx.do_csum = par->do_csum;
	getctx.csum = CSUM_INIT;

	if (!hstor_get(httpstor, par->bucket, key, get_cb, &getctx, false)) {
		fprintf(stderr, "Failed get, bucket %s key %s file %s\n",
			    par->bucket, key, filename);
		/* delete the file? */
		exit(1);
	}

	// if (getctx.off != total) foo

	if (par->do_csum)
		printf("csum 0x%x\n", getctx.csum);
}

static void do_output(struct params *par, struct hstor_client *httpstor, char *errbuf)
{
	int fd;

	if (par->key == NULL) {
		if (!hstor_add_bucket(httpstor, par->bucket)) {
			fprintf(stderr, "Failed to add bucket %s: %s\n",
			    par->bucket, errbuf);
			exit(1);
		}
		return;
	}

	if (isdir(par->file)) {
		output_dir(par, httpstor, par->file, NULL);
		return;
	}

	if (par->file) {
		if ((fd = open(par->file, O_RDONLY)) == -1) {
			fprintf(stderr, "Unable to open %s: %s\n",
			    par->file, strerror(errno));
			exit(1);
		}
		output_one(par, httpstor, par->key, fd, par->file);
		close(fd);
	} else {
		output_one(par, httpstor, par->key, 0, "-");
	}
}

static void output_dir(struct params *par, struct hstor_client *httpstor,
    char *dirname, char *suffix)
{
	DIR *dir;
	struct dirent *ent;
	char *fullname;
	char *newdir;
	char *newsuffix;

	dir = opendir(dirname);
	if (!dir) {
		fprintf(stderr, "Failed to open directory %s: %s\n",
		    dirname, strerror(errno));
		exit(1);
	}

	while ((ent = readdir(dir)) != NULL) {
		if (strcmp(ent->d_name, ".") == 0 ||
		    strcmp(ent->d_name, "..") == 0)
			continue;

		fullname = cat_fs(dirname, ent->d_name);
		if (suffix && suffix[0] != 0) {
			newsuffix = cat_fs(suffix, ent->d_name);
		} else {
			newsuffix = strdup(ent->d_name);
			if (!newsuffix) {
				fprintf(stderr, "No core\n");
				exit(1);
			}
		}

		if (ent->d_type == DT_UNKNOWN) {
			if (isdir(fullname)) {
				newdir = cat_fs(par->file, newsuffix);
				output_dir(par, httpstor, newdir, newsuffix);
				free(newdir);
			} else {
				output_name(par, httpstor, par->key, par->file,
				    newsuffix);
			}
		} else if (ent->d_type == DT_DIR) {
			newdir = cat_fs(par->file, newsuffix);
			output_dir(par, httpstor, newdir, newsuffix);
			free(newdir);
		} else {
			output_name(par, httpstor, par->key, par->file,
			    newsuffix);
		}

		free(fullname);
		free(newsuffix);
	}

	closedir(dir);
}

static void output_name(struct params *par, struct hstor_client *httpstor,
    char *basekey, char *basedir, char *suffix)
{
	int fd;
	char *key;
	char *file;

	key = cat_fs(basekey, suffix);
	file = cat_fs(basedir, suffix);
	if ((fd = open(file, O_RDONLY)) == -1) {
		fprintf(stderr, "Unable to open %s: %s\n",
		    file, strerror(errno));
		exit(1);
	}
	output_one(par, httpstor, key, fd, file);
	close(fd);
	free(key);
	free(file);
}

static void output_one(struct params *par, struct hstor_client *httpstor,
    char *key, int fd, const char *filename)
{
	struct stat statb;
	off_t total;
	struct put_ctx putctx;

	if (fstat(fd, &statb) == -1) {
		fprintf(stderr, "Unable to stat %s: %s\n",
		    filename, strerror(errno));
		exit(1);
	}
	if (!S_ISREG(statb.st_mode)) {
		fprintf(stderr, "Not a regular file '%s'\n", filename);
		exit(1);
	}
	total = statb.st_size;

	memset(&putctx, 0, sizeof(putctx));
	putctx.fd = fd;
	putctx.do_csum = par->do_csum;
	putctx.csum = CSUM_INIT;
	putctx.total = total;

	if (!hstor_put(httpstor, par->bucket, key,
	    put_cb, total, &putctx, NULL)) {
		fprintf(stderr,
		    "Failed to put, bucket %s key %s file %s\n",
		    par->bucket, key, filename);
		exit(1);
	}

	if (putctx.off != total) {
		fprintf(stderr,
		    "Short put, bucket %s key %s file %s off %ld\n",
		    par->bucket, key, filename, (long)putctx.off);
		exit(1);
	}

	if (par->do_csum)
		printf("key %s csum 0x%08x\n", key, putctx.csum);
}

static void do_delete(struct params *par, struct hstor_client *httpstor, char *errbuf)
{

	if (par->key) {
		if (!hstor_del(httpstor, par->bucket, par->key)) {
			fprintf(stderr,
			    "Failed to delete, bucket %s key %s: %s\n",
			    par->bucket, par->key, errbuf);
			exit(1);
		}
	} else {
		if (!hstor_del_bucket(httpstor, par->bucket)) {
			fprintf(stderr, "Failed to delete bucket %s: %s\n",
			    par->bucket, errbuf);
			exit(1);
		}
	}
}

/*
 * N.B. We do not support pagination, so the list is going to be truncated.
 * This is only used for testing for now XXX.
 */
static void do_list(struct params *par, struct hstor_client *httpstor, char *errbuf)
{
	struct hstor_keylist *list;
	struct hstor_object *obj;
	char *pfx, *delim;
	GList *tmp;

	if (!par->bucket) {
		struct hstor_blist *blist;
		struct hstor_bucket *b;

		blist = hstor_list_buckets(httpstor);
		if (!blist) {
			fprintf(stderr, "Failed to list root\n");
			exit(1);
		}

		for (tmp = blist->list; tmp; tmp = tmp->next) {
			b = tmp->data;
			printf("bucket `%s'\n", b->name);
		}

		hstor_free_blist(blist);
		return;
	}

	if (par->key) {
		pfx = par->key;
		delim = "/";
	} else {
		pfx = NULL;
		delim = NULL;
	}
	list = hstor_keys(httpstor, par->bucket, pfx, NULL, delim, 0);
	if (!list) {
		fprintf(stderr, "Failed to list, bucket `%s' prefix `%s'\n",
		    par->bucket, par->key? par->key: "(none)");
		exit(1);
	}

	for (tmp = list->contents; tmp; tmp = tmp->next) {
		obj = tmp->data;
		printf("key `%s' size %ld\n", obj->key, (long)obj->size);
	}

	hstor_free_keylist(list);
}

static size_t put_cb(void *ptr, size_t membsize, size_t nmemb, void *user_data)
{
	struct put_ctx *ctx = user_data;
	size_t len;
	ssize_t rc;

	assert(membsize == 1);

	if (ctx->off >= ctx->total)
		return 0;

	len = nmemb;
	if (len > ctx->total - ctx->off)
		len = ctx->total - ctx->off;
	rc = read(ctx->fd, ptr, len);
	if (rc < 0) {
		fprintf(stderr, "read error: %s", strerror(errno));
		return -1;
	}
	if (rc < len) {
		fprintf(stderr, "short read: %ld at %ld", rc, (long)ctx->off);
		return -1;
	}
	if (ctx->do_csum)
		incrsum(&ctx->csum, ptr, len);

	ctx->off += len;
	return len;
}

static size_t get_cb(void *ptr, size_t membsize, size_t nmemb, void *user_data)
{
	struct get_ctx *ctx = user_data;
	size_t rc;

	assert(membsize == 1);

	if (ctx->do_csum)
		incrsum(&ctx->csum, ptr, nmemb);
	rc = fwrite(ptr, 1, nmemb, ctx->f);
	if (rc < nmemb) {
		if (ferror(ctx->f)) {
			fprintf(stderr, "write error: %s", strerror(errno));
			return -1;
		}
		fprintf(stderr, "short write: %ld at %ld",
		    (long)rc, (long)ctx->off);
		return -1;
	}

	ctx->off += rc;
	return rc;
}

static char *cat_fs(char *dir, char *file)
{
	char *v[3];
	v[0] = dir; v[1] = "/"; v[2] = file;
	return cat(3, v);
}

static char *cat(int n, char *v[])
{
	char *ret;
	size_t len;
	char *p;
	int i;

	if (n == 0)
		return NULL;

	len = 1;
	for (i = 0; i < n; i++)
		len += strlen(v[i]);

	ret = malloc(len);
	if (!ret) {
		fprintf(stderr, "No core\n");
		exit(1);
	}

	p = ret;
	for (i = 0; i < n; i++) {
		len = strlen(v[i]);
		if (len) {
			memcpy(p, v[i], len);
			p += len;
		}
	}
	*p = 0;
	return ret;
}

static void mkpath(char *path)
{
	char *mem;
	char *p;

	mem = strdup(path);
	if (!mem) {
		fprintf(stderr, "No core\n");
		exit(1);
	}

	for (p = mem; *p != 0; p++) {
		if (*p == '/' && p != mem) {
			*p = 0;
			if (mkdir(mem, 0755) == -1) {
				if (errno != EEXIST) {
					fprintf(stderr,
					    "Failed to create directory `%s': %s\n",
					    mem, strerror(errno));
					exit(1);
				}
			}
			*p = '/';
		}
	}

	free(mem);
}

static int isdir(char *filename)
{
	struct stat statb;

	if (filename == NULL)
		return 0;

	if (stat(filename, &statb) == -1) {
		if (errno == ENOENT)
			return 0;
		fprintf(stderr, "Unable to stat %s: %s\n",
		    filename, strerror(errno));
		exit(1);
	}

	if (!S_ISDIR(statb.st_mode))
		return 0;
	return 1;
}

static void incrsum(unsigned int *psum, const unsigned char *data, size_t len)
{
	unsigned int sum;

	sum = *psum;
	while (len) {
		sum ^= *data;
		sum = sum << 1 | sum >> 31;
		data++;
		--len;
	}
	*psum = sum;
}

static void parse_args(struct params *par, int argc, char **argv)
{
	char *arg;

	memset(par, 0, sizeof(struct params));

	++argv;
	while ((arg = *argv++) != NULL) {
		if (arg[0] == '-') {
			switch (arg[1]) {
			case 'b':
				if (*argv == NULL)
					Usage();
				par->bucket = *argv++;
				break;
			case 'c':
				par->do_csum = true;
				break;
			case 'd':
				par->mode = MODE_DEL;
				break;
			case 'f':
				if (*argv == NULL)
					Usage();
				par->file = *argv++;
				break;
			case 'h':
				if (*argv == NULL)
					Usage();
				par->host = *argv++;
				break;
			case 'i':
				par->mode = MODE_IN;
				break;
			case 'k':
				if (*argv == NULL)
					Usage();
				par->key = *argv++;
				break;
			case 'l':
				par->mode = MODE_LIST;
				break;
			case 'n':
				par->subdomain = true;
				break;
			case 'o':
				par->mode = MODE_OUT;
				break;
			case 'p':
				/*
				 * Of course this is a bad idea since other
				 * users can snoop it with ps, but whatever.
				 */
				if (*argv == NULL)
					Usage();
				par->pass = *argv++;
				break;
			case 'u':
				if (*argv == NULL)
					Usage();
				par->user = *argv++;
				break;
			case 'v':
				par->verbose = true;
				break;
			default:
				Usage();
			}
		} else {
			Usage();
		}
	}

	if (par->mode != MODE_IN && par->mode != MODE_OUT &&
	    par->mode != MODE_DEL && par->mode != MODE_LIST) {
		fprintf(stderr, "Mode (-i/-o/-d/-l) is missing\n");
		Usage();
	}
	if (par->host == NULL) {
		fprintf(stderr, "Host (-h) is missing\n");
		Usage();
	}
	if (par->mode != MODE_LIST && par->bucket == NULL) {
		fprintf(stderr, "Bucket (-b) is missing\n");
		Usage();
	}
	if (par->mode == MODE_IN && par->key == NULL) {
		fprintf(stderr, "Key (-k) is missing\n");
		Usage();
	}
	if (par->mode == MODE_OUT && par->key == NULL && par->file != NULL) {
		fprintf(stderr, "Bucket creation needs no file argument\n");
		Usage();
	}
	if (par->mode == MODE_DEL && par->file != NULL) {
		fprintf(stderr, "Deleting needs no file argument\n");
		Usage();
	}
	if (par->mode == MODE_LIST && par->file != NULL) {
		fprintf(stderr, "Listing needs no file argument\n");
		Usage();
	}
	if (par->user == NULL) {
		fprintf(stderr, "User (-u) is missing\n");
		Usage();
	}
	if (par->pass == NULL) {
		fprintf(stderr, "Password (-p) is missing\n");
		Usage();
	}
}

static void Usage(void)
{
	fprintf(stderr, "Usage: poke5 [-i|-o|-d|-l] {options}\n"
			"Modes:\n"
			" -i   Input (from server)\n"
			" -o   Output (to server)\n"
			" -d   Delete\n"
			" -l   List\n"
			"Options:\n"
			" -n            use subdomain for bucket\n"
			" -h host:port  netloc\n"
			" -u user, -p password\n"
			" -b bucket\n"
			" -k key        use filename-like keys, no slashes;"
                                        " empty and none are different\n"
			" -f file       walks directories if pointed to one\n");
	exit(2);
}
